
//
// A virtual listview implementation by Steven E. Sipe
// First published in Dr. Dobb's Journal
//

#include "stdafx.h"
#include "vlistapp.h"
#include "MainDlg.h"
#include "Wait.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CVLDlg dialog

CVLDlg::CVLDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CVLDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CVLDlg)
	m_Search = _T("");
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);

	// Initialize cached data count
	m_nCacheStart = -1;
	m_nCacheEnd = -1;

	// Load the button's bitmap
	m_bmSearch.LoadBitmap(IDB_SEARCH);
}

void CVLDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CVLDlg)
	DDX_Control(pDX, IDC_SEARCHBTN, m_SearchButton);
	DDX_Text(pDX, IDC_SEARCH, m_Search);
	DDX_Control(pDX, IDC_LIST, m_List);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CVLDlg, CDialog)
	//{{AFX_MSG_MAP(CVLDlg)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_WM_DESTROY()
	ON_BN_CLICKED(IDC_SEARCHBTN, OnSearchBtn)
	//}}AFX_MSG_MAP
	ON_MESSAGE(UM_INITIALIZE, OnInitialize)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CVLDlg message handlers

/*
	Creates a table of temporary text items used to test the virtual list 
	control.
*/
void CVLDlg::SetupDB()
{
	CVLDataTable recDataTable(&m_Database);

	try
	{
		// Open the table
		recDataTable.Open();

		// Add some data if there isn't any yet
		if(recDataTable.NumRecords() <= 0)
		{
			const int nMaxItems = 10000;
			CVLWait dlgWait;
			int nCount = 0;
			char szTemp[100];

			// Display the wait dialog and initialize the progress 
			// indicator
			dlgWait.ShowWindow(SW_SHOWNORMAL);
			dlgWait.UpdateWindow();
			dlgWait.SetRange(0,nMaxItems);
			dlgWait.SetStep(1);

			// Iterate until we've reached the maximum number of 
			// entries
			while(nCount < nMaxItems)
			{
				// Create a text string between 5 and 25 characters in length
				for(int j = 0; j < rand()%20+5; j++)
					szTemp[j] = (rand()%26)+'a';
				szTemp[j] = '\0';

				try
				{
					// Add the new item
					recDataTable.AddNew();
					recDataTable.m_Item_Text = szTemp;
					recDataTable.Update();

					nCount++;
					dlgWait.StepIt();
				}
				catch(CDaoException* e)
				{
					// Error 3022 indicates a duplicate key...ignore this error since
					// this is only test code, but consider other errors fatal
					if(e->m_pErrorInfo->m_lErrorCode != 3022)
					{
						AfxMessageBox(e->m_pErrorInfo->m_strDescription);
						e->Delete();
						break;
					}
					e->Delete();
				}
				catch(...)
				{
					AfxMessageBox("Unknown error creating test data");
					break;
				}
			}
		}

		// Close the table
		recDataTable.Close();
	}
	catch(...)
	{
	}
}

BOOL CVLDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Set the search button's bitmap
	m_SearchButton.SetBitmap(m_bmSearch);

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		CString strAboutMenu;
		strAboutMenu.LoadString(IDS_ABOUTBOX);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	try
	{
		// Initialize the list control
		LV_COLUMN lvc;
		lvc.mask = LVCF_FMT|LVCF_SUBITEM|LVCF_TEXT|LVCF_WIDTH;
		lvc.fmt = LVCFMT_LEFT;

		// First column is a row number
		lvc.iSubItem = 0;
		lvc.pszText = "Row";
		lvc.cx = 40;
		m_List.InsertColumn(0,&lvc);

		// The second column contains the item's index number field
		lvc.iSubItem = 1;
		lvc.pszText = "Index";
		lvc.cx = 40;
		m_List.InsertColumn(1,&lvc);

		// The third column contains the item's text field
		lvc.iSubItem = 2;
		lvc.pszText = "Description";
		lvc.cx = 500;
		m_List.InsertColumn(2,&lvc);

		// Initialize the database
		m_Database.Open("data.mdb");

		// Set the data provider
		m_List.SetDataProvider(this);

		// Set an alternating band color, just to make things look less drab
		m_List.SetBand(TRUE,RGB(170,202,240));

		// Finish the initialization...post the message so the 
		// main screen gets painted before we create the database
		// table
		PostMessage(UM_INITIALIZE);
	}
	catch(...)
	{
		// Failed...tell user about it then get out
		AfxMessageBox("Unable to open test data.\n",MB_ICONWARNING|MB_OK);
		PostMessage(WM_CLOSE);
	}

	return TRUE;  
}

LRESULT CVLDlg::OnInitialize(WPARAM,LPARAM)
{
	// Add some data to the test table (if necessary)
	SetupDB();

	// Open the table for use
	m_DataTable.m_strSort = "[Item Text] ASC";
	m_DataTable.Open();

	// Set the number of virtual listbox entries
	int nCount = m_DataTable.NumRecords();
	m_List.SetItemCount(nCount);

	// Just select an arbitrary item
	m_List.SelectItem(123);

	return(0);
}

// Display the "about" dialogbox
void CVLDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.DoModal();
	}
	else
	{
		CDialog::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.
void CVLDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CVLDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}

// Close the database and test table
void CVLDlg::OnDestroy() 
{
	CDialog::OnDestroy();

	m_DataTable.Close();
	m_Database.Close();
}

//////////////////////////////////////
//
// Virtual List Handling Routines
//
//////////////////////////////////////

// Called when the data for a particular item needs to be displayed.  Our
// test list control only has textual information, so that's all we handle.  
// The data to display is first searched for in the cache and then directly
// loaded from the table if it isn't currently cached.
void CVLDlg::OnGetDispInfo(int nID, int iItem, int iSubItem, int mask, int iImage, 
							int state, LPTSTR pszText, int cchTextMax) 
{
	// Is this text data?
	if(mask & LVIF_TEXT)
	{
		int nIndex = iItem;
		DATA_ITEM item;

		// Is this item in the cache already?
		if(nIndex < m_nCacheStart || nIndex > m_nCacheEnd)
		{
			// No, position to the new item
			m_DataTable.SetAbsolutePosition(nIndex);

			// Get its information
			item.nItemNumber = m_DataTable.m_Item_Number;
			item.strItemText = m_DataTable.m_Item_Text;
		}
		else
		{
			// Yes, get the information from the cache
			item = m_arrCachedData[nIndex-m_nCacheStart];
		}

		CString strText;

		// Return the appropariate column of data
		switch(iSubItem)
		{
			// Row  number
			case 0:
				strText.Format("%d",nIndex);
				strncpy(pszText,strText,cchTextMax);
				break;

			// Item index
			case 1:
				strText.Format("%d",item.nItemNumber);
				strncpy(pszText,strText,cchTextMax);
				break;

			// Item text
			case 2:
				strncpy(pszText,item.strItemText,cchTextMax);
				break;
		}
	}
}

// Finds a partial match in the virtual listbox.  This method is used by the 
// OnFindItem() method.  It uses the SQL LIKE operator to search for 
// partial text matches.
int CVLDlg::FindPartial(LPCTSTR lpszItemText)
{
	CVLDataTable recDataTable(&m_Database);
	int nRecNum = -1;

	try
	{
		CString strSQL;

		// Setup a partial SQL match using the LIKE operator
		strSQL.Format("[Item Text] LIKE '%s*'",lpszItemText);

		// Open the query set
		recDataTable.m_strSort = "[Item Text] ASC";
		recDataTable.Open();

		// Did we find a match?
		if(recDataTable.FindFirst(strSQL))
		{
			// Get the index value
			nRecNum = recDataTable.GetAbsolutePosition();
		}

		// Close the table
		recDataTable.Close();
	}
	catch(...)
	{
		AfxMessageBox("Can't find match");
	}

	return(nRecNum);
}

// Handles finding data int the virtual listbox.  This method is called indirectly
// by the CListCtrl::FindItem() method.  Using FindItem() on a virtual listbox
// results in the sending of a LVN_FINDITEM message, which is what this method
// handles.
int CVLDlg::OnFindItem(int nID, LPCTSTR lpszMatch) 
{
	// Get the record number for this entry
	int nIndex = FindPartial(lpszMatch);

	// Was the text invalid?  If so, sound a warning tone
	if(nIndex < 0)
		MessageBeep(0);

	// Return the index of the matching entry (or -1 if it wasn't found)
	return(nIndex);
}

// Handles caching of data items for display.  This method is called to handle the
// LVN_CACHEHIT message.  LVN_CACHEHINT allows the data handler to pre-load 
// the specified data items to allow smooth scrolling among data items.
void CVLDlg::OnCacheHint(int nID, int nFrom, int nTo) 
{
	// Is this range of data outside our current cache?
	if(nFrom != m_nCacheStart)
	{

		// Point to the record for this entry
		m_DataTable.SetAbsolutePosition(nFrom);

		// Load data until we have all of the specified items
		for(int i = 0; i <= nTo-nFrom; i++)
		{
			// Save this record's data
			m_arrCachedData[i].nItemNumber = m_DataTable.m_Item_Number;
			m_arrCachedData[i].strItemText = m_DataTable.m_Item_Text;

			// Get the next record
			m_DataTable.MoveNext();
		}

		// Save the new cache start and finish
		m_nCacheStart = nFrom;
		m_nCacheEnd = nTo;
	}
}

// Handles the search button
void CVLDlg::OnSearchBtn() 
{
	UpdateData(TRUE);

	m_List.SelectItem(m_Search);
}

/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
		// No message handlers
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

