
//
// A virtual listview implementation by Steven E. Sipe
// First published in Dr. Dobb's Journal
//


#include "stdafx.h"
#include "vlist.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CVListCtrl

CVListCtrl::CVListCtrl()
{
	m_pDataInterface = NULL;
	m_nID = -1;
	m_crFG = GetSysColor(COLOR_WINDOWTEXT);
	m_crBG = GetSysColor(COLOR_WINDOW);
	m_crSelBG = GetSysColor(COLOR_HIGHLIGHT);
	m_crSelFG = GetSysColor(COLOR_HIGHLIGHTTEXT);

	m_bBand = FALSE;
	m_crBand = RGB(192,192,192);
}

CVListCtrl::~CVListCtrl()
{
}


BEGIN_MESSAGE_MAP(CVListCtrl, CListCtrl)
	//{{AFX_MSG_MAP(CVListCtrl)
	ON_WM_ERASEBKGND()
	ON_WM_SYSCOLORCHANGE()
	//}}AFX_MSG_MAP
	ON_NOTIFY_REFLECT(LVN_ODFINDITEM, OnLVFindItem)
	ON_NOTIFY_REFLECT(LVN_ODCACHEHINT, OnLVCacheHint)
	ON_NOTIFY_REFLECT(LVN_GETDISPINFO, OnGetDispInfoList)
END_MESSAGE_MAP()

void CVListCtrl::SetColors(COLORREF crFG, COLORREF crBG, COLORREF crSelFG,
								COLORREF crSelBG)
{
	m_crFG = crFG;
	m_crBG = crBG;
	m_crSelFG = crSelFG;
	m_crSelBG = crSelBG;
}

void CVListCtrl::SetBand(BOOL bBand, COLORREF crBand)
{
	m_bBand = bBand;
	m_crBand = crBand;
}

/////////////////////////////////////////////////////////////////////////////
// CVListCtrl message handlers

void CVListCtrl::DrawItem(LPDRAWITEMSTRUCT lpDIS) 
{
	CDC dc;
	dc.Attach(lpDIS->hDC);
	int nCol = 0;
	char szTemp[500];
	LV_COLUMN lvc;
	lvc.mask = LVCF_WIDTH;
	CRect rc = lpDIS->rcItem;
	CRect rcLine = rc;

	BOOL bSelected = GetItemState(lpDIS->itemID,LVIS_SELECTED);
	BOOL bFocused = GetItemState(lpDIS->itemID,LVIS_FOCUSED);
	CPen penGrid(PS_SOLID,1,bSelected?m_crSelFG:m_crSelBG);
	CPen *pOldPen = dc.SelectObject(&penGrid);
	int nOldMode = dc.SetBkMode(TRANSPARENT);

	int nRow = lpDIS->itemID;

	if(m_bBand && nRow && nRow%2)
		dc.FillSolidRect(rc,m_crBand);
	else dc.FillSolidRect(rc,m_crBG);

	COLORREF crText = (bSelected?m_crSelFG:m_crFG);

	// Draw the grid lines and text
	while(TRUE)
	{

		if(GetColumn(nCol,&lvc))
		{
			// Selected??
			if(bSelected)
				dc.FillSolidRect(rc,m_crSelBG);

			dc.MoveTo(rc.left+lvc.cx-1,rc.top);
			dc.LineTo(rc.left+lvc.cx-1,rc.bottom);

			szTemp[0] = '\0';
			GetItemText(lpDIS->itemID,nCol,szTemp,sizeof(szTemp));
			rc.right = rc.left+lvc.cx;

			dc.SetTextColor(crText);

			rc.left += 3;

			int nOffset = 0;
			dc.DrawText(szTemp+nOffset,-1,rc,DT_SINGLELINE|DT_NOPREFIX|
								DT_END_ELLIPSIS);
			rc.left -= 3;

			CRect rcTemp = rcLine;

			rcTemp.left = rc.right;
			rcTemp.right = rc.right;
			rc.left += lvc.cx;
			nCol++;
		}
		else break;
	}

	dc.SetBkMode(nOldMode);
	dc.SelectObject(pOldPen);
	dc.Detach();
}

BOOL CVListCtrl::OnEraseBkgnd(CDC* pDC) 
{
	CRect rc;

	GetClientRect(&rc);
	pDC->FillSolidRect(rc,m_crBG);

	return(TRUE); 
}

void CVListCtrl::SelectItem(int nIndex)
{ 
	CRect rc;

	EnsureVisible(nIndex,FALSE);
	GetItemRect(nIndex,rc,LVIR_BOUNDS);
	SetItemState(nIndex,LVIS_SELECTED,LVIS_SELECTED);
	PostMessage(WM_LBUTTONDOWN,0,MAKELPARAM(rc.left,rc.top));
	PostMessage(WM_LBUTTONUP,0,MAKELPARAM(rc.left,rc.top));
}

void CVListCtrl::SelectItem(LPCTSTR lpszItem)
{
	LV_FINDINFO lvFindInfo;

	lvFindInfo.flags = LVFI_PARTIAL;
	lvFindInfo.psz = lpszItem;
	int nIndex = FindItem(&lvFindInfo);

	if(nIndex >= 0) 
		SelectItem(nIndex);
}

// Tell the virtual listbox how many items we have and that we only want
// to paint the ones that are visible
void CVListCtrl::SetItemCount(int nCount)
{
	SendMessage(LVM_SETITEMCOUNT,nCount,(LPARAM)LVSICF_NOINVALIDATEALL);
}

// Handles finding data int the virtual listbox.  This method is called indirectly
// by the CListCtrl::FindItem() method.  Using FindItem() on a virtual listbox
// results in the sending of a LVN_FINDITEM message, which is what this method
// handles.
void CVListCtrl::OnLVFindItem(NMHDR* pNMHDR, LRESULT* pResult) 
{
	LPNMLVFINDITEM pFindItem = (LPNMLVFINDITEM) pNMHDR;

	int nID = (m_nID >= 0?m_nID:GetDlgCtrlID());

	// Call the data provider to locate the item
	if(m_pDataInterface)
		*pResult = m_pDataInterface->OnFindItem(nID,pFindItem->lvfi.psz);
	else *pResult = 0;
}

// Calls the data provider to handle the caching of data items
void CVListCtrl::OnLVCacheHint(NMHDR* pNMHDR, LRESULT* pResult) 
{
	LPNMLVCACHEHINT pCacheHint = (LPNMLVCACHEHINT) pNMHDR;

	int nID = (m_nID >= 0?m_nID:GetDlgCtrlID());

	// Call the data provider to load the cache
	if(m_pDataInterface)
		m_pDataInterface->OnCacheHint(nID,pCacheHint->iFrom,pCacheHint->iTo);

	*pResult = 0;
}

// Calls data provider to get item data
void CVListCtrl::OnGetDispInfoList(NMHDR* pNMHDR, LRESULT* pResult) 
{
	LV_DISPINFO* pDispInfo = (LV_DISPINFO*)pNMHDR;

	if(m_pDataInterface)
	{
		int nID = (m_nID >= 0?m_nID:GetDlgCtrlID());

		m_pDataInterface->OnGetDispInfo(nID,pDispInfo->item.iItem,
							pDispInfo->item.iSubItem,
							pDispInfo->item.mask, 
							pDispInfo->item.iImage, 
							pDispInfo->item.state, 
							pDispInfo->item.pszText,
							pDispInfo->item.cchTextMax);
	}

	*pResult = 0;
}

// Keep the drawing colors up to date
void CVListCtrl::OnSysColorChange() 
{
	CListCtrl::OnSysColorChange();
	
	m_crFG = GetSysColor(COLOR_WINDOWTEXT);
	m_crBG = GetSysColor(COLOR_WINDOW);
	m_crSelBG = GetSysColor(COLOR_HIGHLIGHT);
	m_crSelFG = GetSysColor(COLOR_HIGHLIGHTTEXT);
}


